from germanium.GermaniumDriver import find_germanium_object


def get_attributes(context, selector):
    """
    Return the attributes for the element that is obtained
    from the selector as a dictionary object.
    :param context:
    :param selector:
    :return:
    """
    germanium = find_germanium_object([context])
    element = germanium.S(selector).element()

    if not element:
        raise Exception("Unable to find '%s' to get_attributes." % selector)

    return germanium.js("""
        var attributes = arguments[0].attributes;
        var result = {};

        for (var i = 0; i < attributes.length; i++) {
            result[attributes[i].name] = attributes[i].value;
        }

        return result;
    """, element)




from .AbstractSelector import AbstractSelector


class InputText(AbstractSelector):
    """
    Just a selector that finds an input text by its name.
    """
    def __init__(self, name=None):
        super(AbstractSelector, self).__init__()
        self._input_name = name

    def get_selectors(self):
        # CSS selector.
        xpath_selectors = ["xpath://input[@type='text']",
                           "xpath://input[not(@type)]"]

        if self._input_name:
            for i in range(len(xpath_selectors)):
                xpath_selectors[i] += "[@name='%s']" % self._input_name

        return xpath_selectors



from time import sleep

import pkg_resources

from .SimpleSelector import SimpleSelector
from .locators import CssLocator, SimpleLocator, XPathLocator, JsLocator
from .iframe_selector import DefaultIFrameSelector, CallableIFrameSelector
from .create_locator import create_locator

from selenium.webdriver.remote.webelement import WebElement


class JavaScriptException(Exception):
    def __init__(self, name, message):
        self.name = name
        self.message = message

    def __str__(self):
        return "JavaScriptException: %s - %s" % (self.name, self.message)


class GermaniumDriver(object):
    """
    A Germanium extension to top of WebDriver
    """
    def __init__(self,
                 web_driver,
                 iframe_selector=DefaultIFrameSelector(),
                 screenshot_folder="screenshots",
                 scripts=list()):

        if hasattr(iframe_selector, '__call__'):
            iframe_selector=CallableIFrameSelector(iframe_selector)

        self.web_driver = web_driver
        self.simple_selector = SimpleSelector(self)
        self._screenshot_folder = screenshot_folder
        self._iframe_selector = iframe_selector
        self._current_iframe = None
        self._scripts_to_load = scripts

        self.locator_map = {
            "simple": SimpleLocator,
            "xpath": XPathLocator,
            "css": CssLocator,
            "js": JsLocator
        }

        self.select_iframe("default")

    def S(self, locator, strategy='detect'):
        """ Finds an element by the given locator. """
        return create_locator(self, locator, strategy)

    def get(self, url, *args, **kwargs):
        result = self.web_driver.get(url, *args, **kwargs)
        self.wait_for_page_to_load()

        return result

    def find_element_by_simple(self, locator):
        """
        Find an element using the simple locator.
        """
        return self.simple_selector.find_element(locator)

    def reload_page(self):
        """
        Reloads the page via JS, and waits for it to load.
        """
        self.js('document.location.reload();')
        self.wait_for_page_to_load()

    def execute_script(self, script, *args, **kwargs):
        """ deprecated js """
        return self.js(script, *args, **kwargs)

    #
    # This executes a script taking care of actually catching and rethrowing correctly
    # JavaScript exceptions.
    #
    # It takes special care for returning web elements directly since if they are exported
    # using a map, under python-3.4 webdriver gets dizzy, and returns the elements as "dict"
    #
    def js(self, script, *args, **kwargs):
        try:
            """
            Execute the script, and also display it on the console for debug purposes.
            """
            wrapper_script = """try {
                var result = {
                    data : (function() {
                        %s
                    }).apply(this, arguments),
                    status : "SUCCESS"
                };

                return result.data && result.data.nodeType && result.data.ownerDocument ?
                       result.data :
                       result;
            } catch (e) {
                return {  // return the exception information in case of failure.
                    status : "FAILURE",
                    name : e.name,
                    message : e.message
                };
            }
            """ % script

            eval_script = wrapper_script

            response = self.web_driver.execute_script(eval_script, *args, **kwargs)

            if isinstance(response, WebElement):
                return response

            if response['status'] == 'SUCCESS':
                return response['data']
            else:
                raise JavaScriptException( response['name'], response['message'] )
        except Exception as e:
            print(wrapper_script)
            print("Failure executing script: %s, error: %s" % (script, e))
            raise e


    def wait_for_page_to_load(self):
        """
        Wait for the page to load.
        """
        self.select_iframe("default")
        self.wait_for_javascript("""return "complete" == document["readyState"]""")

        self.load_simple_locator()  # automatically load the simple locator if waiting for the page to load finished.

        for script_name in self._scripts_to_load:
            self.load_script(script_name)

    def wait_for_action_to_complete(self, throttle=0):
        """
        Waits for all the AJAX and Page calls to finish.
        """
        self.wait_for_ajax_to_complete(throttle)
        self.wait_for_page_to_load()

    def wait_for_ajax_to_complete(self, throttle=0):
        """
        Waits for the action calls to complete.
        """
        self.wait_for_javascript("""window.AJAX_OPERATION_TIMEOUT = %s; return (!XMLHttpRequest.isAjaxRunning || ! XMLHttpRequest.isAjaxRunning()) && "complete" == document["readyState"];""" % throttle)

    def wait_for_javascript(self, script, timeout = 60):
        """
        Executes a script every 400 milliseconds until it returns true. If it goes more than timeout seconds, then this
        function throws an exception. If the given script throws an exception, it's assumed that it returned false.
        """
        start_time = 0 # FIXME: get actual start time.
        wrapper_script = """
            try {
                %s // original script
            } catch (e) {
                console.log(e);
                return false;
            }
        """ % script

        self.wait_for_closure(lambda : self.js(wrapper_script))

    def wait_for_closure(self, closure, timeout = 10):
        """
        Executes a function given as argument every 400 milliseconds until it returns true. If it goes more than
        the timeout seconds, then this function throws an exception. If the function throws an exception, then
        it is assumed it is false.
        """
        def closure_try_catch():
            try:
                return closure()
            except Exception as e:
                print("WARNING: waiting as false since: %s" % e)
                return False

        passed_timeout = 0
        while passed_timeout < timeout and not closure_try_catch():
            passed_timeout += 0.4

            if passed_timeout >= 2:
                print("Running takes more than 2 seconds.")

            sleep(0.4)

    def load_simple_locator(self):
        """
        Since the simple locator script is a bazillion bytes big, it should be loaded independently.
        """
        if self.js("return !window.locateElementBySimple"):
            self.load_script('simple-locator.js')
            self.load_script('ajax-interceptor.js')
            self.load_script('is-ajax-running.js')

    def load_script(self, script_name):
        """
        Load an external script into the current window context.
        """
        script = pkg_resources.resource_string(__name__, script_name)
        if type(script) != 'str':  # it is bytes
            script = script.decode('utf-8')

        self.js(script)

    def take_screenshot(self, name, *args, **kwargs):
        """
        Takes a screenshot of the current browser.
        :param name:
        :return:
        """
        path = self._screenshot_folder + "/" + name + ".png"
        self.web_driver.get_screenshot_as_file(path, *args, **kwargs)
        print("Screenshot saved as " + path)

    def select_iframe(self, iframe_name):
        """
        Selects the iframe, only if the iframe is different
        :param iframe_name:
        :return:
        """
        if iframe_name != self._current_iframe:
            self._current_iframe = self._iframe_selector.select_iframe(self, iframe_name)

        if not self._current_iframe:
            self._current_iframe = iframe_name

        return self._current_iframe

    def __getattr__(self, item):
        """
        Delegate all the attributes that are missing to the web_driver.
        :param item:
        :return:
        """
        if "current_iframe" == item:
            return self._current_iframe

        return getattr(self.web_driver, item)


def find_germanium_object(items):
    """
    Finds the germanium object in the given call. Searches it
    first in the parameter list, and if is not there, it searches
    it in the self reference, as either self.germanium or self._germanium.
    :param items:
    :return:
    """
    if isinstance(items, GermaniumDriver):
        return items

    for item in items:
        if isinstance(item, GermaniumDriver):
            return item

    self = items[0]  # get the self/context reference

    _germanium = getattr(self, "germanium", None)

    if _germanium:
        return _germanium

    return self._germanium



from .AbstractSelector import AbstractSelector


class XPath(AbstractSelector):
    """
    Just a selector that finds some XPath.
    """
    def __init__(self, selector=None):
        self._selector = selector

    def get_selectors(self):
        """ Return the XPath selector itself """
        return ["xpath:" + self._selector]




from .global_germanium_instance import *

from germanium import right_click as right_click_impl


def right_click(selector):
    """
    Right click the element with the given selector.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    right_click_impl(get_instance(), selector)



from selenium.webdriver import ActionChains
from selenium.common.exceptions import NoSuchElementException

from germanium.iframe import find_germanium_object


def click(context, selector = None, move_mouse_over = True):
    """ Click the given selector """
    germanium = find_germanium_object([context])

    element = _element(germanium, selector)
    action = ActionChains(germanium.web_driver)

    if move_mouse_over and selector:
        action.move_to_element(element)

    action.click(element).perform()


def right_click(context, selector = None, move_mouse_over = True):
    """ Right click the given location """
    germanium = find_germanium_object([context])

    element = _element(germanium, selector)
    action = ActionChains(germanium.web_driver)

    if move_mouse_over and selector:
        action.move_to_element(element)

    action.context_click(element).perform()


def double_click(context, selector = None, move_mouse_over = True):
    """ Double click the given location """
    germanium = find_germanium_object([context])

    element = _element(germanium, selector)
    action = ActionChains(germanium.web_driver)

    if move_mouse_over and selector:
        action.move_to_element(element)

    action.double_click(element).perform()


def hover(context, selector = None):
    """ Hover the given location """
    """ Double click the given location """
    germanium = find_germanium_object([context])

    element = _element(germanium, selector)
    action = ActionChains(germanium.web_driver)

    action.move_to_element(element).perform()


def _element(germanium, selector):
    element = None

    if selector:
        element = germanium.S(selector).element()
        if not element:
            raise NoSuchElementException(selector)

    return element





from germanium.selectors import *

# germanium instantiation
from .open_browser import open_browser
from .close_browser import close_browser
from .get_germanium import get_germanium
from .get_web_driver import get_web_driver

# germanium instance functions
from .click import click
from .double_click import double_click
from .go_to import go_to
from .hover import hover
from .right_click import right_click
from .S import S
from .type_keys import type_keys
from .js import js
from .get_attributes import get_attributes

# decorators
from .iframe import iframe


from germanium import click as click_impl
from .global_germanium_instance import *


def click(selector):
    """
    Click the element with the given selector.
    :param selector:
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    click_impl(get_instance(), selector)




class DeferredLocator(object):
    """
    Create a deferred locator that can be used in matching
    elements in wait conditions.
    """
    def __init__(self, germanium):
        self._germanium = germanium

    def __call__(self):
        return self.element()

    def element(self):
        """ Return the current matched element. """
        return self._find_element()

    def element_list(self):
        return self._find_element_list()

    def exists(self):
        """ Return True/False if the currently matched element exists or not """
        return self.element() is not None

    def _find_element(self):
        """ Find the element. """
        raise Exception("not implemented")

    def _find_element_list(self):
        """
        Find all the elements that match the given locator.
        :return:
        """
        raise Exception("Not implemented")



from .global_germanium_instance import *


def go_to(url):
    """
    Go to the given URL, and wait for the page to load.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    get_instance().get(url)



from .DeferredLocator import DeferredLocator

class CompositeLocator(DeferredLocator):
    """
    A locator that will search using the locators it contains.
    """
    def __init__(self, locators):
        super(CompositeLocator, self).__init__(None)

        self._locators = locators

    def _find_element(self):
        for locator in self._locators:
            element = locator.element()
            if element:
                return element
        return None

    def _find_element_list(self):
        result = []

        for locator in self._locators:
            result.extend(locator.element_list())

        return result



from .GermaniumDriver import find_germanium_object

def iframe(target, keep_new_context=False):
    """
    IFrame selector for various operations.
    :param target: The IFrame to set.
    :param keep_new_context: True if the new IFrame should be kept as context after the method is done.
    :param germanium:
    :return:
    """
    def wrapper(original):
        def original_aspect(*args, **kwargs):
            germanium = find_germanium_object(args)
            original_iframe = germanium.current_iframe

            germanium.select_iframe(target)

            try:
                return original(*args, **kwargs)
            finally:
                if not keep_new_context:
                    germanium.select_iframe(original_iframe)

        return original_aspect

    return wrapper



from selenium.common.exceptions import NoSuchElementException

from .DeferredLocator import DeferredLocator


class XPathLocator(DeferredLocator):
    """
    A XPath Deferred locator.
    """
    def __init__(self, germanium, locator):
        super(XPathLocator, self).__init__(germanium)
        self._locator = locator

    def _find_element(self):
        """
        Find an element using the CSS locator provided at creation.
        """
        try:
            return self._germanium.find_element_by_xpath(self._locator)
        except NoSuchElementException as e:
            return None

    def _find_element_list(self):
        """
        Find an element using the CSS locator
        :return:
        """
        try:
            return self._germanium.find_elements_by_xpath(self._locator)
        except NoSuchElementException as e:
            return None



from .DeferredLocator import DeferredLocator


class PositionalFilterLocator(DeferredLocator):
    def __init__(self, locator, left_of_filters=[], right_of_filters=[], above_filters=[], below_filters=[]):
        super(DeferredLocator, self).__init__()
        self.locator = locator
        self.left_of_filters = left_of_filters
        self.right_of_filters = right_of_filters
        self.above_filters = above_filters
        self.below_filters = below_filters

    def _find_element(self):
        items = self._find_element_list()
        if len(items):
            return items[0]

        return None

    def _find_element_list(self):
        elements = self.locator.element_list()

        elements = self._below_filter(elements)
        elements = self._above_filter(elements)
        elements = self._left_of_filter(elements)
        elements = self._right_of_filter(elements)

        return elements

    def _below_filter(self, elements):
        below_elements = []
        for below_filter in self.below_filters:
            below_elements.extend(below_filter.element_list())

        def is_below_all(element):
            for e in below_elements:
                if e.location['y'] <= element.location['y']:
                    return True
            return False

        if len(below_elements):
            elements = list(filter(is_below_all, elements))
            pivot_x = below_elements[0].location['x']
            elements = sorted(elements, key=lambda e: abs(e.location['x'] - pivot_x))
        return elements

    def _above_filter(self, elements):
        above_elements = []
        for above_filter in self.above_filters:
            above_elements.extend(above_filter.element_list())

        def is_above_all(element):
            for e in above_elements:
                if e.location['y'] >= element.location['y']:
                    return True
            return False

        if len(above_elements):
            elements = list(filter(is_above_all, elements))
            pivot_x = above_elements[0].location['x']
            elements = sorted(elements, key=lambda e: abs(e.location['x'] - pivot_x))
        return elements

    def _left_of_filter(self, elements):
        left_of_elements = []
        for left_of_filter in self.left_of_filters:
            left_of_elements.extend(left_of_filter.element_list())

        def is_left_of_all(element):
            for e in left_of_elements:
                if e.location['x'] >= element.location['x']:
                    return True
            return False

        if len(left_of_elements):
            elements = list(filter(is_left_of_all, elements))
            pivot_y = left_of_elements[0].location['y']
            elements = sorted(elements, key=lambda e: abs(e.location['y'] - pivot_y))
        return elements

    def _right_of_filter(self, elements):
        right_of_elements = []
        for right_of_filter in self.right_of_filters:
            right_of_elements.extend(right_of_filter.element_list())

        def is_right_of_all(element):
            for e in right_of_elements:
                if e.location['x'] <= element.location['x']:
                    return True
            return False

        if len(right_of_elements):
            elements = list(filter(is_right_of_all, elements))
            pivot_y = right_of_elements[0].location['y']
            elements = sorted(elements, key=lambda e: abs(e.location['y'] - pivot_y))

        return elements




class DefaultIFrameSelector(object):
    """
    An implementation of the IFrameSelector strategy that does nothing.
    """
    def select_iframe(self, germanium, iframe_name):
        if iframe_name != "default":
            raise Exception("Unknown iframe name: '%s'. Make sure you create an IFrame Selector "
                            "that you will pass when creating the GermaniumDriver, e.g.:\n"
                            "GermaniumDriver(wd, iframe_selector=MyIFrameSelector())")

        germanium.switch_to_default_content()
        return iframe_name


class CallableIFrameSelector(object):
    """
    An implementation of the IFrameSelector strategy that uses the
    given callback and wraps it.
    """
    def __init__(self, callable):
        self._callable = callable

    def select_iframe(self, germanium, iframe_name):
        return self._callable(germanium, iframe_name)



import traceback
import sys


class SimpleSelector(object):
    """
    A SimpleSelector object that is able to find elements
    using the magic of the simple locator.
    """
    def __init__(self, germanium_driver):
        self.germanium_driver = germanium_driver

    def find_element(self, locator):
        """
        Finds several elements using the simple locator.
        Always use two sets of quotes for simple locator arguments! Even for a simple string!!!
        """
        script = "return parent.locateElementBySimple('%s', document);" % (locator.replace("\'", "\\'"))  #escape single quotes

        try:
            return self.germanium_driver.execute_script(script)
        except Exception:  # assume it failed because of the locator
            self.germanium_driver.load_simple_locator()
            try:
                return self.germanium_driver.execute_script(script)
            except Exception as e:
                traceback.print_exc( sys.stdout )
                raise e




class AbstractSelector(object):
    """
    Just a marker interface.
    """
    def __init__(self):
        pass

    def get_selectors(self):
        raise Exception("Abstract class, not implemented.")

    def left_of(self, other_selector):
        return PositionalFilterSelector(self)\
            .left_of(other_selector)

    def right_of(self, other_selector):
        return PositionalFilterSelector(self) \
            .right_of(other_selector)

    def above(self, other_selector):
        return PositionalFilterSelector(self) \
            .above(other_selector)

    def below(self, other_selector):
        return PositionalFilterSelector(self) \
            .below(other_selector)


class PositionalFilterSelector(AbstractSelector):
    """
    Filters selectors
    """
    def __init__(self, parent_selector):
        super(AbstractSelector, self).__init__()

        self.selector = parent_selector

        self.left_of_filters = []
        self.right_of_filters = []
        self.above_filters = []
        self.below_filters = []

    def left_of(self, other_selector):
        self.left_of_filters.append(other_selector)
        return self

    def right_of(self, other_selector):
        self.right_of_filters.append(other_selector)
        return self

    def above(self, other_selector):
        self.above_filters.append(other_selector)
        return self

    def below(self, other_selector):
        self.below_filters.append(other_selector)
        return self




from .DeferredLocator import DeferredLocator

class SimpleLocator(DeferredLocator):
    """
    A Simple Deferred locator.
    """
    def __init__(self, germanium, locator):
        super(SimpleLocator, self).__init__(germanium)
        print("A deprecated SimpleLocator instance was created. SimpleLocator will be removed in a future release.")
        self._locator = locator

    def _find_element_list(self):
        element = self._find_element()

        if element:
            return [element]

        return []

    def _find_element(self):
        """
        Find an element using the Simple locator provided at creation.
        """
        return self._germanium.find_element_by_simple(self._locator)





global_germanium = None


def get_instance():
    """
    :return: GermaniumDriver
    """
    global global_germanium

    return global_germanium


def set_instance(instance):
    global global_germanium

    global_germanium = instance



from germanium import type_keys as type_keys_impl
from .global_germanium_instance import *


def type_keys(keys, selector = None):
    """
    Type the keys specified into the element, or the currently active element.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    type_keys_impl(get_instance(), keys, selector)



from .AbstractSelector import AbstractSelector


class Button(AbstractSelector):
    """
    Just a selector that finds a button by its label or name.
    """
    def __init__(self, search_text=None, text=None, name=None):
        super(AbstractSelector, self).__init__()

        self._search_text = search_text
        self._text = text
        self._name = name

        if self._search_text and self._text:
            raise Exception("You can't have both a searched text and an exact text match")

    def get_selectors(self):
        """ Return the CSS selector to find the button. """
        button_selector = "//button"
        input_selector = "//input[@type='button']"

        if self._name:
            button_selector += "[@name='%s']" % self._name
            input_selector += "[@name='%s']" % self._name

        if self._text:
            button_selector += "[string(.)='%s']" % self._text
            input_selector += "[@value='%s']" % self._text

        if self._search_text:
            button_selector += "[contains(normalize-space(string()), \"%s\")]" % self._search_text
            input_selector += "[contains(@value, \"%s\")]" % self._search_text

        return [input_selector, button_selector]





from germanium.util import create_locator


class SelectorLocator(DeferredLocator):
    """
    A Locator that uses the given selector to locate elements.
    It will defer all the actual work to `create_locator`
    """
    def __init__(self, germanium, selector_instance):
        super(DeferredLocator, self).__init__(germanium)
        self._locator = create_locator(germanium, selector_instance)

    def _findElement(self):
        """
        Finds the element by just using the actual locator that
        was created.
        """
        return self._locator._find_element()

    def _findElements(self):
        """
        Finds the elements using the locator that was created.
        :return:
        """
        return self._locator._find_element_list()


from selenium.webdriver.remote.webelement import WebElement

from .DeferredLocator import DeferredLocator


class JsLocator(DeferredLocator):
    """
    A JS Deferred locator. This locator will execute some JavaScript
    code in order to find the elements.
    """
    def __init__(self, germanium, code):
        super(JsLocator, self).__init__(germanium)

        self._code = code

    def _find_element(self):
        """
        Finds a single element using the provided JS Code.
        """
        element = self._germanium.js(self._code)

        if not element:
            return None

        if not isinstance(element, WebElement):
            raise Exception("Code `%s` is not returning an element." % self._code)

        return element

    def _find_element_list(self):
        """
        Finds a single element using the given code.
        """
        element = self._find_element()

        if not element:
            return []

        return [element]




from time import sleep

def wait(closure, while_not=None, timeout = 10):
    """
    Executes a function given as argument every 400 milliseconds until it returns true. If it goes more than
    the timeout seconds, then this function throws an exception. If the function throws an exception, then
    it is assumed it is false.
    :deprecated:
    """
    def closure_try_catch():
        try:
            return closure()
        except Exception as e:
            print("WARNING: waiting as false since: %s" % e)
            return False

    # compute the while_not_closures that need evaluation.
    while_not_closures = None

    if not while_not:
        while_not_closures = []
    elif hasattr(while_not, '__call__'):
        while_not_closures = [while_not]
    else:
        while_not_closures = while_not

    passed_timeout = 0
    closure_result = False

    while passed_timeout < timeout:
        for while_not_closure in while_not_closures:
            try:
                if while_not_closure():
                    raise Exception("Waiting failed, since while_not condition matched")
            except Exception as e:
                raise Exception("Waiting failed, since while_not condition raised exception", e)

        closure_result = closure_try_catch()

        if closure_result:
            break

        passed_timeout += 0.4

        if passed_timeout >= 2:
            print("Running takes more than 2 seconds.")

        sleep(0.4)

    if not closure_result:
        raise Exception("Waiting has failed")




from .global_germanium_instance import *


def get_web_driver():
    """
    Returns the currently used web_driver object by germanium.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    return get_instance().web_driver



from .AbstractSelector import AbstractSelector


class Css(AbstractSelector):
    """
    Just a selector that finds some CSS.
    """
    def __init__(self, selector=None):
        self._selector = selector

    def get_selectors(self):
        """ Return the CSS selector itself """
        return ["css:" + self._selector]





from .CompositeLocator import CompositeLocator
from .CssLocator import CssLocator
from .DeferredLocator import DeferredLocator
from .SimpleLocator import SimpleLocator
from .StaticElementLocator import StaticElementLocator
from .XPathLocator import XPathLocator
from .JsLocator import JsLocator

from .PositionalFilterLocator import PositionalFilterLocator



from germanium import hover as hover_impl
from .global_germanium_instance import *


def hover(selector):
    """
    Hover the element with the given selector.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    hover_impl(get_instance(), selector)



from .AbstractSelector import AbstractSelector, PositionalFilterSelector

from .Css import Css
from .XPath import XPath
from .JsSelector import JsSelector

from .Button import Button
from .Element import Element
from .Input import Input
from .InputText import InputText
from .Link import Link
from .Text import Text



import re

from selenium import webdriver

from germanium import GermaniumDriver
from germanium.GermaniumDriver import DefaultIFrameSelector
from .global_germanium_instance import *


REMOTE_ADDRESS = re.compile(r"^(\w+?):(.*?)$")


def open_browser(browser="Firefox",
                 wd=None,
                 iframe_selector=DefaultIFrameSelector(),
                 screenshot_folder="screenshots",
                 scripts=list()):
    """
    Open the given browser.
    :param browser:
    :param wd:
    :param iframe_selector:
    :param screenshot_folder:
    :param scripts:
    """
    if get_instance():
        raise Exception("A browser already runs. Close it first with close_browser()")

    remote_match = REMOTE_ADDRESS.match(browser)

    if wd:
        web_driver = wd
    elif remote_match:
        remote_browser = remote_match.group(1)

        if remote_browser.lower() == "firefox" or remote_browser.lower() == "ff":
            remote_capabilities = webdriver.DesiredCapabilities.FIREFOX
        elif remote_browser.lower() == "chrome":
            remote_capabilities = webdriver.DesiredCapabilities.CHROME
        elif remote_browser.lower() == "ie":
            remote_capabilities = webdriver.DesiredCapabilities.INTERNETEXPLORER
        else:
            raise Exception("Unknown browser: %s, only firefox, "
                            "chrome and ie are supported." % browser)

        web_driver = webdriver.Remote(command_executor=remote_match.group(2),
                                      desired_capabilities=remote_capabilities)
    elif browser.lower() == "firefox" or browser.lower() == "ff":
        web_driver = webdriver.Firefox()
    elif browser.lower() == "chrome":
        web_driver = webdriver.Chrome()
    elif browser.lower() == "ie":
        web_driver = webdriver.Ie()
    else:
        raise Exception("Unknown browser: %s, only firefox, "
                        "chrome and ie are supported." % browser)

    set_instance(GermaniumDriver(web_driver,
                                 iframe_selector=iframe_selector,
                                 screenshot_folder=screenshot_folder,
                                 scripts=scripts))

    return get_instance()



from .AbstractSelector import AbstractSelector


class Element(AbstractSelector):
    def __init__(self,
                 tag_name=None,
                 index=-1,
                 exact_text=None,
                 contains_text=None,
                 css_classes=[],
                 exact_attributes={}):
        """ A general element selector. """
        self._tag_name = tag_name
        self._index = index
        self._text = exact_text
        self._contains_text = contains_text
        self._css_classes = css_classes
        self._exact_attributes = exact_attributes

        if self._text and self._contains_text:
            raise Exception("Having the exact text to be matched, "
                            "and a partial text to be searched is not supported.")

    def get_selectors(self):
        xpath_locator = '//' + self._tag_name

        if self._contains_text:
            xpath_locator += "[contains(normalize-space(string()), '%s')]" % self._contains_text

        if self._text:
            xpath_locator += "[string() = '%s']" % self._text

        for css_class in self._css_classes:
            xpath_locator += '[contains(concat(" ", @class, " "), " %s ")]' % css_class

        for k, v in self._exact_attributes.items():
            xpath_locator += '[@%s = "%s"]' % (k, v)

        if self._index >= 0:
            xpath_locator = '(%s)[%d]' % (xpath_locator, self._index)

        return ["xpath:" + xpath_locator]



from .AbstractSelector import AbstractSelector


class Link(AbstractSelector):
    """
    A selector that finds links in the page.
    """
    def __init__(self, search_text=None, text=None, search_href=None, href=None):
        super(AbstractSelector, self).__init__()

        self._search_text = search_text
        self._text = text
        self._search_href = search_href
        self._href = href

    def get_selectors(self):
        """ Return the xpsth locator to find the text """
        if self._search_text and self._text:
            raise Exception("You can't have both a searched text and an exact text match")

        if self._search_href and self._href:
            raise Exception("You can't have both a searched href and an exact href match")

        css_locator = 'xpath://a'

        if self._search_text:
            css_locator += "[contains(normalize-space(string()), \"%s\")]" % self._search_text

        if self._text:
            css_locator += '[string()="%s"]' % self._text

        if self._search_href:
            css_locator += "[contains(@href, \"%s\")]" % self._search_href

        if self._href:
            css_locator += '[@href="%s"]' % self._href

        return [css_locator]




from .AbstractSelector import AbstractSelector


class JsSelector(AbstractSelector):
    """
    Just a selector that finds some elements using some
    Js code. The code must return a single WebDriver element.
    """
    def __init__(self, code=None):
        super(JsSelector, self).__init__()

        self._code = code

    def get_selectors(self):
        """ Return the JavaScript selector itself """
        return ["js:" + self._code]



from .global_germanium_instance import get_instance


def iframe(target, keep_new_context = False):
    """
    Switch the iframe in static contexts.
    """
    def wrapper(original):
        def original_aspect(*args, **kwargs):
            germanium = get_instance()
            original_iframe = germanium.current_iframe

            germanium.select_iframe(target)

            try:
                return original(*args, **kwargs)
            finally:
                if not keep_new_context:
                    germanium.select_iframe(original_iframe)

        return original_aspect

    return wrapper




from .AbstractSelector import AbstractSelector


class Input(AbstractSelector):
    """
    Just a selector that finds an input by its name.
    """
    def __init__(self, name=None):
        super(AbstractSelector, self).__init__()
        self._input_name = name

    def get_selectors(self):
        # CSS selector.
        css_selector = "input"

        if self._input_name:
            css_selector += "[name='%s']" % self._input_name

        return [css_selector]



from .AbstractSelector import AbstractSelector


class Text(AbstractSelector):
    """
    Just a selector that finds the text in the page.
    """
    def __init__(self, text):
        super(AbstractSelector, self).__init__()
        self._text = text

    def get_selectors(self):
        """ Return the simple selector to find the text """
        js_selector = """js:
            var searchedText = "%s";

            function text(node) {
                return node.innerText || node.textContent;
            }

            if (text(document.body).indexOf(searchedText) < 0) {
                return null;
            }

            var result = document.body;
            var foundChildElement = true;

            while (foundChildElement) {
                foundChildElement = false;

                for (var i = 0; i < result.children.length; i++) {
                    if (text(result.children[i]).indexOf(searchedText) >= 0) {
                        foundChildElement = true;
                        result = result.children[i];
                    }
                }
            }

            return result;
        """ % str(self._text).replace('"', '\\"', 100000)

        return [js_selector]



from .global_germanium_instance import *


def S(*argv, **kwargs):
    """
    Call the super selector from germanium.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    return get_instance().S(*argv, **kwargs)



from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='germanium',
    version='1.6.3',
    description='The germanium project: Selenium WebDriver testing API that doesn\'t disappoint.',
    long_description = readme,
    author='Bogdan Mustiata',
    author_email='bogdan.mustiata@gmail.com',
    license='BSD',

    install_requires=['selenium'],
    packages=['germanium', 'germanium.locators', 'germanium.selectors', 'germanium.static', 'germanium.util'],
    package_data={'germanium': ['*.js']}
)



from .global_germanium_instance import *


def close_browser():
    """
    Close the currently running browser.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    get_instance().quit()
    set_instance(None)



from germanium.util import get_attributes as get_attributes_impl

from .global_germanium_instance import get_instance


def get_attributes(selector):
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    return get_attributes_impl(get_instance(), selector)



from .DeferredLocator import DeferredLocator


class StaticElementLocator(DeferredLocator):
    def __init__(self, element):
        """ Just holds a static reference to a single element. """
        super(StaticElementLocator, self).__init__(None)

        self._element = element

    def _find_element(self):
        """ Returns the locally stored element. """
        return self._element

    def _find_element_list(self):
        return [self._element]



from .global_germanium_instance import *


def js(code, *args, **kwargs):
    """
    Right click the element with the given selector.
    """
    if not get_instance():
        raise Exception("You need to start a browser first with open_browser()")

    return get_instance().js(code, *args, **kwargs)



import re

from selenium.webdriver.remote.webelement import WebElement

from germanium.locators import \
    XPathLocator, \
    CssLocator, \
    SimpleLocator, \
    CompositeLocator, \
    DeferredLocator, \
    StaticElementLocator, \
    PositionalFilterLocator
from germanium.selectors import AbstractSelector, PositionalFilterSelector

LOCATOR_SPECIFIER = re.compile(r'((\w[\w\d]*?)\:)(.*)', re.MULTILINE|re.DOTALL)


def create_locator(germanium, locator, strategy='detect'):
    if strategy == 'css':
        return CssLocator(germanium, locator)

    if strategy == 'xpath':
        return XPathLocator(germanium, locator)

    if strategy == 'simple':
        return SimpleLocator(germanium, locator)

    if strategy != 'detect':
        locator_constructor = germanium.locator_map[strategy]

        if not locator_constructor:
            raise Exception('Unable to find strategy %s. Available strategies: detect, %s' % (strategy, ', '.join(germanium.locator_map.keys())))

        return locator_constructor(germanium, locator)

    if isinstance(locator, DeferredLocator):
        if strategy is not 'detect':
            raise Exception('The locator is already constructed, but a strategy is also defined: "%s"' % strategy)

        return locator

    if isinstance(locator, PositionalFilterSelector):
        left_of_filters = map(lambda x: create_locator(germanium, x),
                              locator.left_of_filters)

        right_of_filters = map(lambda x: create_locator(germanium, x),
                               locator.right_of_filters)

        above_filters = map(lambda x: create_locator(germanium, x),
                            locator.above_filters)

        below_filters = map(lambda x: create_locator(germanium, x),
                            locator.below_filters)

        return PositionalFilterLocator(
            locator=create_locator(germanium, locator.selector),
            left_of_filters=left_of_filters,
            right_of_filters=right_of_filters,
            above_filters=above_filters,
            below_filters=below_filters
        )

    if isinstance(locator, AbstractSelector):
        selectors = locator.get_selectors()

        # if there is only one locator, don't apply the composite.
        if len(selectors) == 1:
            return create_locator(germanium, selectors[0])

        # if we have multiple locators, apply the composite locator.
        locator_list = []
        for selector in locator.get_selectors():
            locator_list.append(create_locator(germanium, selector))

        return CompositeLocator(locator_list)

    if isinstance(locator, WebElement):
        return StaticElementLocator(locator)

    # if it starts with // it's probably an XPath locator.
    if locator[0:2] == "//":
        return XPathLocator(germanium, locator)

    m = LOCATOR_SPECIFIER.match(locator)
    if m:
        locator_constructor = germanium.locator_map[m.group(2)]
        if locator_constructor:
            return locator_constructor(germanium, m.group(3))

    return CssLocator(germanium, locator)

